#! /bin/bash -e


# Copyright (c) 2007 Takahiro Shinozaki

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# Synopsis
#   train-tied-state-triphone.bash [-debug] [-config config] [-doneflg doneflg]

# Description:
#    train tied state triphone HMM

# Change log:
# Thu Dec 27 16:29:13 JST 2007
# The initial version was written by Takahiro Shinozaki.

[ -n "$debug" ] && set -x
scriptp=$0
script=`basename $0`

function usage {
    echo "Usage: $script [-debug]"
    echo "          [-config run.config]"
    echo "          [-doneflg DONE_FLAG]"
    exit 1
}

# initialize local variables
doneflg=done
envlog=env.log

while [ -n "$1" ]; do
    case "$1" in
        -debug) shift; export debug=1; set -x;;
        -config) shift; config="$1"; shift; . $config;;
        *) usage;
    esac
done

if [ -f $doneflg ]; then
    echo "Nothing to be done ($script)."
    exit 0
fi

# dump environmental variables
env > $envlog

# print process info
echo --$script START--
date
pwd
if [ ! -e $script ]; then
    cp -p $scriptp .
fi

# Set default values
export initthr=1000

# All the input global variables used in this script are (should be) listed here.
echo HCompV=$HCompV	# HTK command
echo HERest=$HERest	# HTK command
echo HHEd=$HHEd		# HTK command
echo scp=$scp		# training data list file
echo proto0=$proto0	# prototype HMM
echo monomlf=$monomlf	# monophone master label file
echo trimlf=$trimlf	# triphone master label file
echo monohlist=$monohlist	# monophone hlist
echo corpustri=$corpustri	# corpus triphone list
echo alltri=$alltri	# (physical) triphone list
                        # (Using logical list is ok but redundant HMMs are not reduced)
echo trtiehed=$trtiehed	# HHEd editor command file
echo nparallel=$nparallel	# number of parallel processes
echo niter=$niter	# number of EM iteration at each training stage
echo beam=$beam		# beam width
echo nstates=$nstates	# number of HMM states
echo nmixes=$nmixes	# number of mixtures
echo tdctemplate=$tdctemplate	# HHEd editor command meta file
echo brun=$brun		# batch execution script
echo bruncfg=$bruncfg	# batch execution config
echo bruncfgL=$bruncfgL	# batch execution config (for memory intensive process)
echo bruncfgD=$bruncfgD	# batch execution config (for disk access intensive process)
echo brunLsize=$brunLsize	# HMM size threshold to use bruncfgL
echo rmiaccs=$rmiaccs	# rm acc files during training (to save disk)
echo accfext=$accfext	# accumulator file extension
echo initthr=$initthr   # initial threshold for tree clustering

# check requirements
if [ "$HCompV" = "" ]; then echo HCompV is not defined; exit 1; fi
if [ "$HERest" = "" ]; then echo HERest is not defined; exit 1; fi
if [ "$HHEd" = "" ]; then echo HHEd is not defined; exit 1; fi
if [ "$scp" = "" ]; then echo scp is not defined; exit 1; fi
if [ "$proto0" = "" ]; then echo proto0 is not defined; exit 1; fi
if [ "$monomlf" = "" ]; then echo monomlf is not defined; exit 1; fi
if [ "$trimlf" = "" ]; then echo trimlf is not defined; exit 1; fi
if [ "$monohlist" = "" ]; then echo monohlist is not defined; exit 1; fi
if [ "$corpustri" = "" ]; then echo corpustri is not defined; exit 1; fi
if [ "$alltri" = "" ]; then echo alltri is not defined; exit 1; fi
if [ "$trtiehed" = "" ]; then echo trtiehed is not defined; exit 1; fi
if [ "$nparallel" = "" ]; then echo nparallel is not defined; exit 1; fi
if [ "$niter" = "" ]; then echo niter is not defined; exit 1; fi
if [ "$beam" = "" ]; then echo beam is not defined; exit 1; fi
if [ "$nstates" = "" ]; then echo nstates is not defined; exit 1; fi
if [ "$nmixes" = "" ]; then echo nmixes is not defined; exit 1; fi
if [ "$tdctemplate" = "" ]; then echo tdctemplate is not defined; exit 1; fi
if [ "$brun" = "" ]; then echo brun is not defined; exit 1; fi
# if [ "$initthr" = "" ]; then echo initthr is not defined; exit 1; fi

# initialize model parameters
if [ ! -f proto/done ]; then
    mkdir -p proto
    (cd proto; export outdir=.; proto.bash)
else
    echo skip proto
fi
if [ ! -f proto/done ]; then echo proto failed; exit 1; fi

# subdivide scp
mkdir -p sdscpd
subdivide.pl $scp $nparallel sdscpd scp

# train monophone parameters
if [ ! -f mono/done ]; then
    mkdir -p mono
    (
	cd mono

	proto=../proto/proto
	vfloor=../proto/vFloors
	if [ ! -f iter-0/done ]; then
            mkdir -p iter-0
            proto2mono.pl $proto $vfloor $monohlist > iter-0/hmmdefs
	fi

	export hlist=$monohlist
	export mlf=$monomlf
	export hmm0=iter-0/hmmdefs
	export sdscpd=../sdscpd
	pem.bash
    )
else
    echo skip mono
fi
if [ ! -f mono/done ]; then echo mono failed; exit 1; fi

# train triphone parameters
if [ ! -f tri/done ]; then
    mkdir -p tri
    (
	cd tri

	monohmm=../mono/iter-$niter/hmmdefs
	if [ ! -f iter-0/done ]; then
            mkdir -p iter-0
	    echo "CL $corpustri" > mktrihed
	    cat $trtiehed >> mktrihed
            $HHEd -H $monohmm -M iter-0 mktrihed $monohlist
            touch iter-0/done
	fi

	export mlf=$trimlf
	export hlist=$corpustri
	export hmm0=iter-0/hmmdefs
	export sdscpd=../sdscpd
	pem.bash
    )
else
    echo skip tri
fi
if [ ! -f tri/done ]; then echo tri failed; exit 1; fi

# train Gaussian mixture tied-state HMM
if [ ! -f tri-st$nstates/done ]; then
    mkdir -p tri-st$nstates
    (
	cd tri-st$nstates

	# HMM state clustering
	if [ ! -f cluster/done ]; then
            mkdir -p cluster
	    (
		cd cluster

		export orghmm=../../tri/iter-$niter/hmmdefs
		export stats=../../tri/iter-$niter/stats
		htkclustering.bash
		mv env.log tdc.env
		if [ ! -f done ]; then echo clustering failed; exit 1; fi
	     )
	else
            echo skip clustering
	fi

	# Increase the number of mixtures
	if [ ! -f mix/done ]; then
            mkdir -p mix
	    (
		cd mix

		export mlf=$trimlf
		export hlist=$alltri
		export sdscpd=../../../sdscpd

		mix=1
		while [ $mix -le $nmixes ]; do
		    if [ -f mix-$mix/done ]; then
			echo skip mix-$mix.
			mix=$(( mix * 2 ))
			continue;
		    fi

		    mkdir -p mix-$mix

		    pmix=$(( $mix / 2 ))
		    if [ ! -f mix-$mix/iter-0/done ]; then
			mkdir -p mix-$mix/iter-0
			if [ $pmix -eq 0 ]; then
			    orghmm=../cluster/st$nstates/hmmdefs
			    cp -f $orghmm mix-$mix/iter-0
			else
			    orghmm=mix-$pmix/iter-$niter/hmmdefs
			    echo "MU $mix  {*.state[2-4].mix}" > mixup$mix.hed
			    $HHEd -T 1 -H $orghmm -M mix-$mix/iter-0 mixup$mix.hed $hlist
			fi
			if [ -s mix-$mix/iter-0/hmmdefs ]; then
			    touch mix-$mix/iter-0/done
			else
			    echo "failed to initialize mix-$mix"
			    exit 1
			fi
		    fi

		    # update model parameters
		    (cd mix-$mix; export hmm0=iter-0/hmmdefs; pem.bash)

		    # check
		    if [ ! -f mix-$mix/done ]; then
			echo failed to train mix-$mix
			exit 1
		    fi
		    mix=$(( mix * 2 ))
		done
		touch done
	    )
	else
            echo skip mix
	fi
	if [ ! -f mix/done ]; then
	    echo failed to train mix
	    exit 1
	fi
	touch done
    )
else
    echo skip tistmix
fi
if [ ! -f tri-st$nstates/done ]; then
    echo failed to train tri-st$nstates
    exit 1
fi

echo done $script
touch done
echo --$script END--
